// =============================================================================
// Module com_ram_tdp_async_reg
// Revision 20170109a : Initial revision
// Asynchronous true dual type memory by registers.
// This module can be compiled by Verilog-HDL 2001.
// =============================================================================

`timescale 1ps/1ps

module com_ram_tdp_async_reg (
    clk_a,
    din_a,
    addr_a,
    en_a,
    wen_a,
    dout_a,

    clk_b,
    din_b,
    addr_b,
    en_b,
    wen_b,
    dout_b
);

    // parameter ---------------------------------------------------------------
    parameter BIT_WIDTH = 16;
    parameter ADR_WIDTH = 3;

    // local parameter ---------------------------------------------------------
    localparam ADR_DEPTH = 1 << ADR_WIDTH;
    localparam UD = 1;

    // input/output ------------------------------------------------------------
    input                     clk_a;
    input  [BIT_WIDTH-1:0]    din_a;
    input  [ADR_WIDTH-1:0]    addr_a;
    input                     en_a;
    input                     wen_a;
    output [BIT_WIDTH-1:0]    dout_a;
    
    input                     clk_b;
    input  [BIT_WIDTH-1:0]    din_b;
    input  [ADR_WIDTH-1:0]    addr_b;
    input                     en_b;
    input                     wen_b;
    output [BIT_WIDTH-1:0]    dout_b;

    // write -------------------------------------------------------------------
    // * Only clock path is figured. Data path is not described.
    //
    //          clk_a_wen[0]-->+---------+             +-----------+
    //          clk_b_wen[0]-->| i_cg[0] |             |   mem[0]  |--->
    // clk_a--+--------------->|         |-clk_mem[0]->|>          |
    // clk_b-+---------------->+---------+             +-----------+
    //       ||
    //       || clk_a_wen[1]-->+---------+             +-----------+
    //       || clk_b_wen[1]-->| i_cg[1] |             |   mem[1]  |--->
    //       |+--------------->|         |-clk_mem[1]->|>          |
    //       +---------------->+---------+             +-----------+
    //       ||       :             :                        :
    //       ||       :             :                        :
    //       || clk_a_wen[n]-->+---------+             +-----------+
    //       || clk_b_wen[n]-->| i_cg[n] |             |   mem[n]  |--->
    //       |+--------------->|         |-clk_mem[n]->|>          |
    //       +---------------->+---------+             +-----------+

    wire   [ADR_DEPTH-1:0]    clk_a_wen;
    wire   [ADR_DEPTH-1:0]    clk_b_wen;
    wire   [ADR_DEPTH-1:0]    clk_mem;

    reg    [BIT_WIDTH-1:0]    mem [ADR_DEPTH-1:0];

    generate
        genvar i;
        for (i=0; i<ADR_DEPTH; i=i+1)
        begin: tdp_reg_ab

            assign clk_a_wen[i] = (addr_a == i) & en_a & wen_a;
            assign clk_b_wen[i] = (addr_b == i) & en_b & wen_b;

            com_ram_tdp_clkgate_ab i_cg (
                .clk_a(clk_a),
                .clk_b(clk_b),
                .clk_a_en(clk_a_wen[i]),
                .clk_b_en(clk_b_wen[i]),
                .clk_out(clk_mem[i])
            );

          // synopsys translate_off
            always @(clk_a_wen[i] or clk_b_wen[i]) begin
                if (clk_a_wen[i] & clk_b_wen[i]) begin
                    $display("[WARNING] %m : Same write address(%0d) from A/B port at %0t[ps]"
                             ,i ,$stime);
                end
            end
          // synopsys translate_on

            always @(posedge clk_mem[i]) begin
                mem[i] <= #UD
                  // synopsys translate_off
                    (clk_a_wen[i] & clk_b_wen[i])? {BIT_WIDTH{1'bx}} :
                  // synopsys translate_on
                    (clk_a_wen[i])? din_a :
                    (clk_b_wen[i])? din_b : mem[i];
            end

        end
    endgenerate

    // read --------------------------------------------------------------------
    reg    [ADR_WIDTH-1:0] r_addr_a;
    reg    [ADR_WIDTH-1:0] r_addr_b;
    wire   [ADR_WIDTH-1:0] w_addr_a;
    wire   [ADR_WIDTH-1:0] w_addr_b;

    // synopsys translate_off
      always @(en_a or wen_a or en_b or wen_b or addr_a or addr_b) begin
          if (en_a & ~wen_a & en_b & wen_b & (addr_a == addr_b)) begin
              $display("[WARNING] %m : Same read/write address(%0d) from A/B port at %0t[ps]"
                       ,addr_a ,$stime);
          end
          else
          if (en_b & ~wen_b & en_a & wen_a & (addr_a == addr_b)) begin
              $display("[WARNING] %m : Same write/read address(%0d) from A/B port at %0t[ps]"
                       ,addr_b ,$stime);
          end
      end
    // synopsys translate_on

    assign w_addr_a = 
      // synopsys translate_off
        (en_a & ~wen_a & en_b & wen_b & (addr_a == addr_b))? {ADR_WIDTH{1'bx}} :
      // synopsys translate_on
        (en_a & ~wen_a)? addr_a : r_addr_a;

    assign w_addr_b = 
      // synopsys translate_off
        (en_b & ~wen_b & en_a & wen_a & (addr_a == addr_b))? {ADR_WIDTH{1'bx}} :
      // synopsys translate_on
        (en_b & ~wen_b)? addr_b : r_addr_b;

    always @(posedge clk_a) begin
        r_addr_a <= #UD w_addr_a;
    end
    always @(posedge clk_b) begin
        r_addr_b <= #UD w_addr_b;
    end

    assign dout_a = mem[r_addr_a];
    assign dout_b = mem[r_addr_b];

endmodule
